/*
** Module   : FrameWidget.java
** Abstract : server side frame widget class (contains the configuration)
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23491 Created initial version.
** 002 NVS 20060114   @24097 Frame is created hidden.
** 003 SIY 20060213   @24581 Fixes for the DialogBox support.
** 004 SIY 20060426   @25911 Changed signatures of get/setNextTabItem.
** 005 SIY 20060616   @27537 Added isDown(),setName() methods.
** 006 GES 20060803   @28377 Some cleanup and refactoring.
** 007 ECF 20061207   @31598 Fixed setAttrSpace(). Was calling the wrong
**                           method in FrameConfig.
** 008 CA  20090413   @41775 Fixes for FRAME:HIDDEN attribute. On init,
**                           HIDDEN flag is off. The setter must change
**                           the HIDDEN flag, not VISIBLE.
** 009 GES 20101005          Added a non-functional COM-HANDLE attribute stub.
** 010 SIY 20101006          Added moveAfterTab() stub.
** 011 SVL 20101006          #10 moved to GenericWidget.
** 012 SIY 20101007          Added setName() with different signature.
** 013 SIY 20111101          Updated imports.
** 014 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 015 CA  20130927          Added LegacyResource annotation, to determine resource type.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * This is the server side frame widget class which contains the 
 * client-side configuration for this frame BUT which does NOT implement
 * any of the server side logic for frame processing.
 * <p>
 * WARNING: unlike most of the server-side widget classes, this class
 * intentionally DOES NOT push a modified screen definition to the client
 * when visibility/hidden/down changes occur.  Such changes transfer to the
 * client using alternate paths and since these are frequently changed, using
 * the screen definition for such things is not acceptable.
 */
@LegacyResource(resource = LegacyResource.FRAME)
public class FrameWidget
extends ContainerEntity
{
   /** Reference to our configuration which is pre-cast to the proper type. */
   private FrameConfig fcfg = null;
   
   /**
    * Default constructor.
    */
   public FrameWidget()
   {
      super(new FrameConfig(ComponentConfig.FRAME, null));
      fcfg = (FrameConfig) config;
      config.setHidden(false);
   }

   /**
    * Set NAME attribute.
    * 
    * @param   name
    *          New value for the attribute.
    */
   public void setName(String name)
   {
      fcfg.setName(name);
   }
   
   /**
    * Set NAME attribute.
    * 
    * @param   name
    *          New value for the attribute.
    */
   public void setName(character name)
   {
      setName(name.isUnknown() ? "" : name.toStringMessage());
   }
   
   /**
    * Gets the NEXT-TAB-ITEM writable attribute.
    *
    * @return   The current value of NEXT-TAB-ITEM attribute.
    */
   public GenericWidget getNextTabItem()
   {
      // TODO: implement this!
      throw new RuntimeException("NEXT-TAB-ITEM is not done yet!");
   }

   /**
    * Sets the NEXT-TAB-ITEM writable attribute.
    *
    * @param    n
    *           The new value for the NEXT-TAB-ITEM attribute.
    */
   public void setNextTabItem(GenericWidget n)
   {
      // TODO: implement this!
      throw new RuntimeException("NEXT-TAB-ITEM is not done yet!");
   }

   /**
    * Gets the ATTR-SPACE writable attribute.
    *
    * @return   The current value of the ATTR-SPACE attribute.
    */
   public logical isAttrSpace()
   {
      return new logical(fcfg.isAttrSpace());
   }

   /**
    * Sets the ATTR-SPACE writable attribute.
    *
    * @param    space
    *           The new value for the ATTR-SPACE attribute.
    */
   public void setAttrSpace(logical space)
   {
      setAttrSpace(space.booleanValue());
   }

   /**
    * Sets the ATTR-SPACE writable attribute.
    *
    * @param    space
    *           The new value for the ATTR-SPACE attribute.
    */
   public void setAttrSpace(boolean space)
   {
      fcfg.setAttrSpace(space);
      pushScreenDefinition();
   }
   
   /**
    * Sets the dialog box flag.
    *
    * @param dialogBox
    *        new value for the dialog box flag
    */
   public void setDialogBox(boolean dialogBox)
   {
      fcfg.setDialog(dialogBox);
   }
   
   /**
    * Gets the dialog box flag.
    *
    * @return <code>true</code> if this frame is a dialog box.
    */
   public boolean isDialogBox()
   {
      return fcfg.isDialog();
   }

   /**
    * Sets the down frame size.
    *
    * @param    down
    *           The new number of iterations for this down frame.
    */
   public void setDown(int down)
   {
      fcfg.setDown(down);
   }
   
   /**
    * Gets the down frame size.
    *
    * @return   The number of data iterations for the frame (1 if the
    *           frame is not a down frame).
    */
   public int getDown()
   {
      return fcfg.getDown();
   }

   /**
    * Return <code>true</code> if frame is DOWN frame.
    * 
    * @return   <code>true</code> if frame is DOWN frame.
    */
   public boolean isDown()
   {
      return (getDown() != 1);
   }
   
   /**
    * Sets the new value for HIDDEN flag.
    * 
    * @param   hidden
    *          The new boolean value for the HIDDEN attribute.
    */
   public void setHidden(boolean hidden)
   {
      config.setHidden(hidden);
   }

   /**
    * Sets the Progress VISIBLE attribute.
    * 
    * @param   visible
    *          The new value for the Progress VISIBLE attribute.
    */
   public void setVisible(boolean visible)
   {
      config.setVisible(visible);
   }
   
   /**
    * Gets the Progress COM-HANDLE attribute.
    * <p>
    * This is a non-functional stub for now.
    * 
    * @return   The value for the COM-HANDLE attribute.
    */
   public comhandle getComHandle()
   {
      return new comhandle();
   }
}
