/*
** Module   : XNodeRefImpl
** Abstract : DOM XML Node reference resource type implementation.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 EVL 20130204 Created initial version. Major rework to use w3c backend.
** 002 CA  20130221 Added LOCAL-NAME support. 
** 003 EVL 20130222 Adding NODE-VALUE attribute and SET-ATTRIBUTE-NODE method.
** 004 EVK 20130531 Adding NODE-VALUE attribute and GET-PARENT, GET-ATTRIBUTE-NODE, NORMALIZE
**                  LONGCHAR-TO-NODE-VALUE, NODE-VALUE-TO-LONGCHAR, CLONE-NODE methods.
** 005 CA  20130930 Changes related to determining resource type from LegacyResource annotation.
*/

package com.goldencode.p2j.xml;

import org.w3c.dom.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of the X-Node-Reference specific features of the Progress
 * DOM XML support objects.
 */
public class XNodeRefImpl
extends XEntityImpl
implements XNodeRef
{

   public static final String RESOURCE_TYPE = "X-NODEREF";

   /**
    * Default constructor.
    */
   public XNodeRefImpl()
   {
      // no-op
   }

   /**
    * Special constructor accepting the namespace URI specification to make
    * another node reference case.
    *
    * @param    node
    *           The node associated with this node.
    */
   public XNodeRefImpl(Node node)
   {
      super(node);
   }

   // Attributes
   /**
    * Getting the attribute names available for current node reference.
    * 
    * @return   The comma separated character string of all available
    *           attributed.
    */
   @Override
   public character getAttributeNames()
   {
      if (!isInitialized())
      {
         invalidNodeError("ATTRIBUTE-NAMES");
         return new character();
      }

      StringBuilder sb = new StringBuilder();
      NamedNodeMap nnm = getNode().getAttributes();
      if (nnm == null)
      {
         return new character("");
      }
      int iSize = nnm.getLength();
      for (int i = 0; i < iSize; i++)
      {
         sb.append(((Attr)nnm.item(i)).getName());
         if (i < iSize - 1)
         {
            sb.append(',');
         }
      }   

      return new character(sb.toString());
   }

   /**
    * Getting the value of the XML node value attribute.
    * 
    * @return   The character value of the node reference.
    */
   @Override
   public character getNodeValue()
   {
      if (!isInitialized())
      {
         invalidNodeError("NODE-VALUE");
         return new character();
      }
      try
      {
         String nodeValue = getNode().getNodeValue();
         return new character(nodeValue == null ? "" : nodeValue);
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.DOMSTRING_SIZE_ERR)
         {
            commonDomError("NODE-VALUE", "Too long node-value");
         }
         return new character();
      }
   }
   
   /**
    * Setting the value of the XML node value attribute.
    * 
    * @param    value
    *           The new value of the node value attribute.
    */
   @Override
   public void setNodeValue(character value)
   {
      if (value.isUnknown())
      {
         invalidArgumentAssignError("NODE-VALUE");
         return;
      }
      if (!isInitialized())
      {
         String[] errMsg =
            {
               "X-NODEREF must be associated with a valid X-DOCUMENT in order to use it in " +
                  "method NODE-VALUE",
               "Unable to set attribute NODE-VALUE in widget  of type X-NODEREF"
            };
         int[] errorCodes = { 9102, 3131 };
         ErrorManager.recordOrShowError(errorCodes, errMsg, false, false, true);
         return;
      }

      Node node = getNode();
      int nodeType = node.getNodeType();
      if (nodeType == Node.ELEMENT_NODE || nodeType == Node.ENTITY_REFERENCE_NODE)
      {
         String[] errMsg =
            {
               "X-NODEREF or X-DOCUMENT NODE-VALUE got an error: An " +
                  "attempt was made to modify an object where modifications are not allowed",
               "Unable to set attribute NODE-VALUE in widget  of type X-NODEREF"
            };
         int[] errorCodes = { 9082, 3131 };
         ErrorManager.recordOrShowError(errorCodes, errMsg, false, false, true);
         return;
      }

      try
      {
         node.setNodeValue(value.toStringMessage());
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
         {
            String[] errMsg =
               {
                  "X-NODEREF or X-DOCUMENT NODE-VALUE got an error: An " +
                     "attempt was made to modify an object where modifications are not allowed",
                  "Unable to set attribute NODE-VALUE in widget  of type X-NODEREF"
               };
            int[] errorCodes = { 9082, 3131 };
            ErrorManager.recordOrShowError(errorCodes, errMsg, false, false, true);
            return;
         }
      }
   }
   
   /**
    * Setting the value of the XML node value attribute.
    * 
    * @param    value
    *           The new value of the node value attribute.
    */
   @Override
   public void setNodeValue(String value)
   {
      setNodeValue(new character(value));
   }

   /**
    * Creating the new node within the current Document object.
    *
    * @param xNodeRef The valid handle class instance to use for the cloned XML node.
    * @param deep     A logical that if <code>true</code> specifies that the whole sub-trees is to be
    *                 copied. The default value is {@code false}.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical cloneNode(handle xNodeRef, logical deep)
   {
      if (!isInitialized())
      {
         invalidNodeError("CLONE-NODE");
         return new logical(false);
      }

      if (!isXNodeHandle(xNodeRef))
      {
         invalidArgumentError("CLONE-NODE");
         return new logical(false);
      }

      Node node = getNode();
      Node clonedNode = node.cloneNode(deep.booleanValue());
      setNodeToHandle(xNodeRef, clonedNode);

      return xNodeRef.isValid();
   }

   /**
    * Creating the new node within the current Document object.
    *
    * @param xNodeRef The valid handle class instance to use for the cloned XML node.
    * @param deep     A logical that if <code>true</code> specifies that the whole sub-trees is to
    *                 be copied. The default value is {@code false}.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical cloneNode(handle xNodeRef, boolean deep)
   {
      return cloneNode(xNodeRef, new logical(deep));
   }

   /**
    * Getting the Parent of current node.
    *
    * @param    hNoderefParent
    *           A valid handle to assign parent XML node.
    *
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical getParent(handle hNoderefParent)
   {
      if (!isInitialized())
      {
         invalidNodeError("GET-PARENT");
         return new logical(false);
      }
      if (!isXNodeHandle(hNoderefParent))
      {
         invalidArgumentError("GET-PARENT");
         return new logical(false);
      }

      Node node = getNode().getParentNode();
      setNodeToHandle(hNoderefParent, node);

      return hNoderefParent.isValid();
   }

   // Methods
   /**
    * Deleting the internals of the current X-Node-Reference.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical deleteNode()
   {
      if (!isInitialized())
      {
         invalidNodeError("DELETE-NODE");
         return new logical(false);
      }

      Node node = getNode();
      Node parentNode = node.getParentNode();
      if (parentNode == null)
      {
         invalidNodeError("DELETE-NODE");
         return new logical(false);
      }
      try
      {
         parentNode.removeChild(node);
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.NOT_FOUND_ERR)
         {
            invalidNodeError("DELETE-NODE");
         }
         else if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
         {
            invalidNodeError("DELETE-NODE");
         }
         return new logical(false);
      }

      return new logical(true);
   }
   
   /**
    * Getting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to return.
    *
    * @return   The comma separated character string of all available
    *           attributed.
    */
   @Override
   public character getAttribute(String name)
   {
      return getAttribute(new character(name));
   }
   
   /**
    * Getting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to return.
    *
    * @return   Value as a character string, or the empty string
    *           if that attribute does not have a specified or default value.
    */
   @Override
   public character getAttribute(character name)
   {
      if (!isInitialized())
      {
         invalidNodeError("GET-ATTRIBUTE");
         return new character();
      }
      if (name.isUnknown())
      {
         invalidArgumentError("GET-ATTRIBUTE");
         return new character();
      }

      Node node = getNode();
      String result = null;
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
         Element elementNode = (Element) node;
         result = elementNode.getAttribute(name.toStringMessage());
      }

      return new character((result == null) ? "" : result);
   }
   
   /**
    * Setting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to change.
    * @param    value
    *           The new value for the given attribute name.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical setAttribute(String name, String value)
   {
      return setAttribute(new character(name), new character(value));
   }
   
   /**
    * Setting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to change.
    * @param    value
    *           The new value for the given attribute name.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical setAttribute(String name, character value)
   {
      return setAttribute(new character(name), value);
   }
   
   /**
    * Setting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to change.
    * @param    value
    *           The new value for the given attribute name.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical setAttribute(character name, String value)
   {
      return setAttribute(name, new character(value));
   }
   
   /**
    * Setting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to change.
    * @param    value
    *           The new value for the given attribute name.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical setAttribute(character name, character value)
   {
      if (!isInitialized())
      {
         invalidNodeError("SET-ATTRIBUTE");
         return new logical(false);
      }
      if (name.isUnknown() || value.isUnknown())
      {
         invalidArgumentError("SET-ATTRIBUTE");
         return new logical(false);
      }

      Node node = getNode();
      final logical result;
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
         Element elementNode = (Element) node;
         try
         {
            elementNode.setAttribute(name.toStringMessage(), value.toStringMessage());
         }
         catch (DOMException e)
         {
            if (e.code == DOMException.INVALID_CHARACTER_ERR)
            {
               commonDomError("SET-ATTRIBUTE", "An invalid character was specified");
            }
            else if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
            {
               commonDomError("SET-ATTRIBUTE", "Read only node");
            }
            return new logical(false);
         }
         result = new logical(true);
      }
      else
      {
         commonDomError("SET-ATTRIBUTE", "Invalid node type");
         result = new logical(false);
      }

      return result;
   }

   /**
    * Associates XML attribute node with the given X-Noderef object handle.
    * 
    * @param    handle
    *           The handle that represents an XML attribute node created with create node or
    *           create node namespace.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical setAttributeNode(handle handle)
   {
      if (handle._isValid() && !handle.isUnknown() && handle.get() instanceof HandleResource)
      {
         HandleResource resource = (HandleResource) handle.get();
         if (LegacyResource.X_DOCUMENT.equalsIgnoreCase(resource.type()))
         {
            invalidArgumentError("SET-ATTRIBUTE-NODE");
            return new logical(false);
         }
      }

      if (!isInitialized() || handle.isUnknown())
      {
         invalidNodeError("SET-ATTRIBUTE-NODE");
         return new logical(false);
      }

      if (!isXNodeHandle(handle))
      {
         invalidArgumentError("SET-ATTRIBUTE-NODE");
         return new logical(false);
      }
      Node currentNode = getNode();
      Object resource = handle.get();
      Node handleNode = ((XNodeRefImpl) resource).getNode();
      final logical result;
      if (currentNode.getNodeType() == Node.ELEMENT_NODE &&
          handleNode.getNodeType() == Node.ATTRIBUTE_NODE)
      {
         Element elementNode = (Element) currentNode;
         try
         {
            elementNode.setAttributeNode((Attr) handleNode);
         }
         catch (DOMException e)
         {
            if (e.code == DOMException.WRONG_DOCUMENT_ERR)
            {
               invalidNodeError("SET-ATTRIBUTE-NODE");
            }
            else if (e.code == DOMException.INUSE_ATTRIBUTE_ERR)
            {
               commonDomError("SET-ATTRIBUTE-NODE", "Attribute already exists");
            }
            else if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
            {
               commonDomError("SET-ATTRIBUTE-NODE",
                              "An attempt was made to modify an object " +
                              "where modifications are not allowed");
            }
            return new logical(false);
         }

         result = new logical(true);
      }
      else
      {
         commonDomError("SET-ATTRIBUTE-NODE", "Invalid node type");
         result = new logical(false);
      }

      return result;
   }

   /**
    * Associates XML attribute node with the specified name to the given X-Noderef object
    * handle.
    *
    * @param    handle
    *           The handle that represents an XML attribute node with the specified name.
    * @param    attributeName
    *           The string that represents an XML attribute node with the specified name.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical getAttributeNode(handle handle, String attributeName)
   {
      return getAttributeNode(handle, new character(attributeName));
   }

   /**
    * Associates XML attribute node with the specified name to the given X-Noderef object
    * handle.
    *
    * @param    handle
    *           The handle that represents an XML attribute node with the specified name.
    * @param    attributeName
    *           The character that represents an XML attribute node with the specified name.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical getAttributeNode(handle handle, character attributeName)
   {
      if (handle._isValid() && !handle.isUnknown() && handle.get() instanceof HandleResource)
      {
         HandleResource resource = (HandleResource) handle.get();
         if (LegacyResource.X_DOCUMENT.equalsIgnoreCase(resource.type()))
         {
            invalidArgumentError("GET-ATTRIBUTE-NODE");
            return new logical(false);
         }
      }

      if (!isXNodeHandle(handle) || attributeName.isUnknown())
      {
         invalidArgumentError("GET-ATTRIBUTE-NODE");
         return new logical(false);
      }

      if (!isInitialized())
      {
         invalidNodeError("GET-ATTRIBUTE-NODE");
         return new logical(false);
      }

      Node node = getNode();
      final logical result;
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
         Element elementNode = (Element) node;
         Attr attrNode = elementNode.getAttributeNode(attributeName.toStringMessage());
         if (attrNode != null)
         {
            setNodeToHandle(handle, attrNode);
         }
         result = handle.isValid();
      }
      else
      {
         commonDomError("GET-ATTRIBUTE-NODE", "Invalid node type");
         result = new logical(false);
      }
      return result;
   }

   /**
    * Setting the binary value of the XML node value attribute.
    *
    * @param    value
    *           The new value of the node value attribute.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical longcharToNodeValue(longchar value)
   {
      if (!isInitialized())
      {
         invalidNodeError("LONGCHAR-TO-NODE-VALUE");
         return new logical(false);
      }

      if (value == null || value.isUnknown())
      {
         invalidArgumentError("LONGCHAR-TO-NODE-VALUE");
         return new logical(false);
      }

      Node node = getNode();
      node.setTextContent(value.toStringMessage());

      return new logical(true);
   }

   /**
    * Setting the XML node value attribute to binary value of longchar.
    *
    * @param    value
    *           To save the value on XML node to this parameter.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical nodeValueToLongchar(longchar value)
   {
      if (!isInitialized())
      {
         invalidNodeError("NODE-VALUE-TO-LONGCHAR");
         return new logical(false);
      }

      value.setValue(getNode().getNodeValue());

      return new logical(true);
   }

   /**
    * Setting the binary value of the XML node value attribute.
    *
    * @param    value
    *           The new value of the node value attribute.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical memptrToNodeValue(memptr value)
   {
      if (!isInitialized())
      {
         invalidNodeError("MEMPTR-TO-NODE-VALUE");
         return new logical(false);
      }

      if (value == null || value.isUnknown())
      {
         invalidArgumentError("MEMPTR-TO-NODE-VALUE");
      }

      Node node = getNode();
      try
      {
         node.setTextContent(new String(value.getByteArray()));
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
         {
            commonDomError("MEMPTR-TO-NODE-VALUE", "Read only node");//TODO need correct message
            return new logical(false);
         }
      }
      return new logical(true);
   }

   /**
    * Setting the XML node value attribute to binary value of memptr.
    *
    * @param    value
    *           To save the value on XML node to this parameter.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical nodeValueToMemptr(memptr value)
   {
      if (!isInitialized())
      {
         invalidNodeError("NODE-VALUE-TO-MEMPTR");
         return new logical(false);
      }

      value.setString(getNode().getNodeValue(), 1);

      return new logical(true);
   }

   /**
    * Normalizes TEXT and ATTRIBUTE nodes in the full depth of the sub-tree under this
    * XML node.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical normalize()
   {
      if (!isInitialized())
      {
         invalidNodeError("NORMALIZE");
         return new logical(false);
      }

      Node node = getNode();
      short nodeType = node.getNodeType();
      if (nodeType == Node.ELEMENT_NODE)
      {
         node.normalize();
      }
      else
      {
         commonDomError("NORMALIZE", "Invalid node type");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Get the value of the LOCAL-NAME attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getLocalName()
   {
      if (!isInitialized())
      {
         invalidNodeError("LOCAL-NAME");
         return new character();
      }

      String localName = getNode().getLocalName();
      if (localName != null)
      {
         return new character(localName);
      }
      else
      {
         return new character("");
      }
   }

   /**
    * Setter for NAMESPACE-PREFIX attribute.
    *
    * @param    prefix
    *           The new value of the NAMESPACE-PREFIX attribute.
    */
   @Override
   public void setNamespacePrefix(character prefix)
   {
      if (!isInitialized())
      {
         String[] errMsg =
         {
            "X-NODEREF must be associated with a valid X-DOCUMENT in order " +
               "to use it in method NAMESPACE-PREFIX",
            "Unable to set attribute NAMESPACE-PREFIX in widget  of type X-NODEREF"
         };
         int[] errorCodes = { 9102, 3131 };
         ErrorManager.recordOrShowError(errorCodes, errMsg, false, false, true);
         return;
      }

      Node currentNode = getNode();
      String prefixStr = prefix.getValue(); //null is ok here
      try
      {
         if (currentNode.getNodeType() == Node.ELEMENT_NODE)
         {
            Element elementNode = (Element) currentNode;
            elementNode.setPrefix(prefixStr);
         }
         else if (currentNode.getNodeType() == Node.ATTRIBUTE_NODE)
         {
            Attr attributeNode = (Attr) currentNode;
            attributeNode.setPrefix(prefixStr);
         }
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
         {
            commonDomError("NAMESPACE-PREFIX",
                           "An attempt was made to modify an object where " +
                           "modifications are not allowed");
         }
         else if (e.code == DOMException.NAMESPACE_ERR ||
                  e.code == DOMException.INVALID_CHARACTER_ERR)
         {
            String[] errMsg =
            {
               "X-NODEREF or X-DOCUMENT NAMESPACE-PREFIX got an error: " +
                  "XML Parser not initialized",
               "Unable to set attribute NAMESPACE-PREFIX in widget  of type X-NODEREF"
            };
            int[] errorCodes = { 9082, 3131 };
            ErrorManager.recordOrShowError(errorCodes, errMsg, false, false, true);
         }
      }
   }

   /**
    * Gets the number of child for the current one. Only for XNodRef
    *
    * @return   The integer number of the child, 1-based index of the child node.
    */
   @Override
   public integer getChildNum()
   {
      if (!isInitialized())
      {
         invalidNodeError("NUM-CHILDREN");
         return new integer();
      }

      if (LegacyResource.X_DOCUMENT.equalsIgnoreCase(type()))
      {
         return new integer();
      }

      Node node = getNode();
      Node parent = node.getParentNode();
      if (parent == null)
      {
         return new integer(1);
      }

      NodeList childNodes = parent.getChildNodes();
      for (int i = 0 ; i < childNodes.getLength(); i++)
      {
         if (node.equals(childNodes.item(i)))
         {
            return new integer(i + 1);
         }
      }

      return new integer();
   }

   /**
    * Getting the Document the current reference belongs to.
    *
    * @return   The handle of the Document containing tne current reference.
    */
   @Override
   public handle getOwnerDocument()
   {
      if (!isInitialized())
      {
         invalidNodeError("OWNER-DOCUMENT");
         return new handle();
      }

      return new handle(new XNodeRefImpl(getNode().getOwnerDocument()));
   }
}
